import asyncio
import collections
import datetime
import functools
import logging
import os
import re
from typing import Optional

import aiohttp
import bs4
import pandas as pd
import requests
from pytse_client import (
    config,
    symbols_data,
    translations,
    tse_settings,
    utils,
)
from pytse_client.download import download, download_ticker_client_types_record
from pytse_client.scraper import tsetmc_scraper
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL
from pytse_client.utils import async_utils
from tenacity import retry, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)
logger.addHandler(logging.NullHandler())
RealtimeTickerInfo = collections.namedtuple(
    'RealtimeTickerInfo', [
        'last_price',
        'adj_close',
        'best_demand_vol',
        'best_demand_price',
        'best_supply_vol',
        'best_supply_price',
    ]
)


class Ticker:
    def __init__(self, symbol: str, index: Optional[str] = None):
        self.symbol = symbol
        self.csv_path = f"{config.DATA_BASE_PATH}/{self.symbol}.csv"
        self._index = index or symbols_data.get_ticker_index(self.symbol)
        self._url = tse_settings.TSE_TICKER_ADDRESS.format(self._index)
        self._info_url = tse_settings.TSE_ISNT_INFO_URL.format(self._index)
        self._client_types_url = TSE_CLIENT_TYPE_DATA_URL.format(self._index)
        self._history: pd.DataFrame = pd.DataFrame()

        if os.path.exists(self.csv_path):
            self.from_file()
        else:
            self.from_web()

    def from_web(self):
        self._history = download(self._index)[self._index]

    def from_file(self):
        self._history = pd.read_csv(self.csv_path)
        self._history["date"] = pd.to_datetime(self._history["date"])

    @property
    def history(self):
        return self._history

    @property
    def url(self):
        return self._url

    @property
    def index(self):
        return self._index

    @property
    def instrument_id(self):
        """
        instrument id of a ticker is unique and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"InstrumentID='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def ci_sin(self):
        """
        instrument id of a ticker is like instrument_id and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"CIsin='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def title(self) -> str:
        return re.findall(r"Title='(.*?)',", self._ticker_page_response.text
                          )[0].split("-")[0].strip()

    @property
    def group_name(self) -> str:
        return re.findall(
            r"LSecVal='([\D]*)',", self._ticker_page_response.text
        )[0]

    @property
    def p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have p/e
        """
        adj_close = self.get_ticker_real_time_info_response().adj_close
        eps = self.eps
        if adj_close is None or eps is None or eps == 0:
            return None
        return self.get_ticker_real_time_info_response().adj_close / self.eps

    @property
    def group_p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like وملت does not have group P/E (gpe)
        """
        gpe = re.findall(
            r"SectorPE='([\d.]*)',", self._ticker_page_response.text
        )
        if not gpe:
            return None
        return float(gpe[0])

    @property
    def eps(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have eps
        """
        eps = re.findall(
            r"EstimatedEPS='([-,\d]*)',", self._ticker_page_response.text
        )[0]
        if eps == "":
            return None
        return float(eps)

    @property
    def total_shares(self) -> float:
        return float(
            re.findall(r"ZTitad=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def base_volume(self) -> float:
        return float(
            re.findall(r"BaseVol=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def client_types(self):
        return download_ticker_client_types_record(self._index)

    @property
    def trade_dates(self):
        return self._history["date"].to_list()

    @property
    def shareholders(self) -> pd.DataFrame:
        session = utils.requests_retry_session()
        page = session.get(self._shareholders_url, timeout=5)
        session.close()
        soup = bs4.BeautifulSoup(page.content, 'html.parser')
        table: bs4.PageElement = soup.find_all("table")[0]
        shareholders_df = utils.get_shareholders_html_table_as_csv(table)
        shareholders_df = shareholders_df.rename(
            columns=translations.SHAREHOLDERS_FIELD_MAPPINGS
        )
        return shareholders_df

    def get_shareholders_history(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None
    ) -> pd.DataFrame:
        """
            a helper function to use shareholders_history_async
        """
        return asyncio.run(
            self.get_shareholders_history_async(
                from_when,
                to_when,
                only_trade_days,
                session,
            ),
        )

    async def get_shareholders_history_async(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None,
    ) -> pd.DataFrame:
        requested_dates = utils.datetime_range(to_when - from_when, to_when)
        session_created = False
        if not session:
            session_created = True
            conn = aiohttp.TCPConnector(limit=3)
            session = aiohttp.ClientSession(connector=conn)
        tasks = []
        for date in requested_dates:
            if only_trade_days and date.date() not in self.trade_dates:
                continue
            tasks.append(
                self._get_ticker_daily_info_page_response(
                    session, date.strftime(tse_settings.DATE_FORMAT)
                )
            )
        pages = await async_utils.run_tasks_with_wait(tasks, 30, 10)
        if session_created is True:
            await session.close()
        rows = []
        for page in pages:
            page_date = tsetmc_scraper.scrape_daily_info_page_for_date(page)
            shareholders_data = (
                tsetmc_scraper.
                scrape_daily_info_page_for_shareholder_data(page)
            )
            for shareholder_data in shareholders_data:
                rows.append(
                    [
                        datetime.datetime.strptime(
                            page_date,
                            tse_settings.DATE_FORMAT,
                        ),
                        shareholder_data.shares,
                        shareholder_data.percentage,
                        shareholder_data.instrument_id,
                        shareholder_data.name,
                    ]
                )

        return pd.DataFrame(
            data=rows,
            columns=[
                'date',
                'shares',
                'percentage',
                'instrument_id',
                'shareholder',
            ]
        )

    @property
    def last_price(self):
        return self.get_ticker_real_time_info_response().last_price

    @property
    def adj_close(self):
        return self.get_ticker_real_time_info_response().adj_close

    @property
    def best_demand_vol(self):
        return self.get_ticker_real_time_info_response().best_demand_vol

    @property
    def best_demand_price(self):
        return self.get_ticker_real_time_info_response().best_demand_price

    @property
    def best_supply_vol(self):
        return self.get_ticker_real_time_info_response().best_supply_vol

    @property
    def best_supply_price(self):
        return self.get_ticker_real_time_info_response().best_supply_price

    def get_ticker_real_time_info_response(self) -> RealtimeTickerInfo:
        """
        notes on usage:
        - Real time data might not be always available
        check for None values before usage
        """
        session = utils.requests_retry_session()
        response = session.get(self._info_url, timeout=5)
        session.close()
        # check supply and demand data exists
        if response.text.split(";")[2] != "":
            best_demand_vol = int(response.text.split(";")[2].split("@")[1])
            best_demand_price = int(response.text.split(";")[2].split("@")[2])
            best_supply_vol = int(response.text.split(";")[2].split("@")[4])
            best_supply_price = int(response.text.split(";")[2].split("@")[3])
        else:
            best_demand_vol = None
            best_demand_price = None
            best_supply_vol = None
            best_supply_price = None

        # in some cases last price or adj price is undefined
        try:
            last_price = int(response.text.split()[1].split(",")[1])
        except (ValueError, IndexError):  # When instead of number value is `F`
            last_price = None
        try:
            adj_close = int(response.text.split()[1].split(",")[2])
        except (ValueError, IndexError):
            adj_close = None
        return RealtimeTickerInfo(
            last_price,
            adj_close,
            best_demand_vol=best_demand_vol,
            best_demand_price=best_demand_price,
            best_supply_vol=best_supply_vol,
            best_supply_price=best_supply_price,
        )

    @property
    @functools.lru_cache()
    def _ticker_page_response(self):
        return utils.requests_retry_session().get(self._url, timeout=10)

    @functools.lru_cache()
    @retry(
        wait=wait_random(min=3, max=5),
        before_sleep=before_sleep_log(logger, logging.ERROR)
    )
    async def _get_ticker_daily_info_page_response(
        self, session, date
    ) -> requests.Response:
        async with session.get(
            tse_settings.INSTRUMENT_DAY_INFO_URL.format(
                index=self.index, date=date
            ),
        ) as response:
            response.raise_for_status()
            page = await response.text()
            logger.info(f"fetched date {date}")
            return page

    @property
    def _shareholders_url(self) -> str:
        return tse_settings.TSE_SHAREHOLDERS_URL.format(self.ci_sin)



import datetime
from time import sleep

from pytse_client import Ticker

symbols = [
    "شکبیر",
    "برکت",
]
from_times = [
    datetime.timedelta(days=90),
    datetime.timedelta(days=150),
]
for symbol in symbols:
    print(f"downloading {symbol}")
    Ticker(symbol).get_shareholders_history().to_csv(f"{symbol}.csv")
    print(f"downloaded {symbol} complete")
    # sleep between to avoid getting blocked
    sleep(50)



from concurrent import futures
from io import StringIO
from pathlib import Path
from typing import Dict, List, Union

import pandas as pd
from requests import HTTPError
import jdatetime

from pytse_client import config, symbols_data, translations, tse_settings
from pytse_client.utils import requests_retry_session
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL


def download(
        symbols: Union[List, str],
        write_to_csv: bool = False,
        include_jdate: bool = False,
        base_path: str = config.DATA_BASE_PATH) -> Dict[str, pd.DataFrame]:
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        session = requests_retry_session()
        for symbol in symbols:
            if symbol.isnumeric():
                ticker_index = symbol
            else:
                ticker_index = symbols_data.get_ticker_index(symbol)
                _handle_ticker_index(symbol, ticker_index)
            future = executor.submit(
                download_ticker_daily_record, ticker_index, session
            )
            future_to_symbol[future] = symbol
        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            df: pd.DataFrame = future.result()
            df = df.iloc[::-1]
            df = df.rename(columns=translations.HISTORY_FIELD_MAPPINGS)
            df = df.drop(columns=["<PER>", "<TICKER>"])
            _adjust_data_frame(df, include_jdate)
            df_list[symbol] = df
            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                df.to_csv(f'{base_path}/{symbol}.csv')

    if len(df_list) != len(symbols):
        print("Warning, download did not complete, re-run the code")
    session.close()
    return df_list


def _adjust_data_frame(df, include_jdate):
    df.date = pd.to_datetime(df.date, format="%Y%m%d")
    if include_jdate:
        df['jdate'] = ""
        df.jdate = df.date.apply(
            lambda gregorian:
            jdatetime.date.fromgregorian(date=gregorian))


def download_ticker_daily_record(ticker_index: str, session):
    url = tse_settings.TSE_TICKER_EXPORT_DATA_ADDRESS.format(ticker_index)
    response = session.get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        return download_ticker_daily_record(ticker_index, session)

    data = StringIO(response.text)
    return pd.read_csv(data)


def download_client_types_records(
        symbols: Union[List, str],
        write_to_csv: bool = False,
        include_jdate: bool = False,
        base_path: str = config.CLIENT_TYPES_DATA_BASE_PATH
):
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        for symbol in symbols:
            ticker_index = symbols_data.get_ticker_index(symbol)
            _handle_ticker_index(symbol, ticker_index)
            future = executor.submit(
                download_ticker_client_types_record,
                ticker_index
            )
            future_to_symbol[future] = symbol
        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            df: pd.DataFrame = future.result()
            _adjust_data_frame(df, include_jdate)
            df_list[symbol] = df
            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                df.to_csv(f'{base_path}/{symbol}.csv')
    return df_list


def _handle_ticker_index(symbol, ticker_index):
    if ticker_index is None:
        ticker_index = get_symbol_id(symbol)
        if ticker_index is None:
            raise Exception("Can not found ticker name")
        else:
            symbols_data.append_symbol_to_file(ticker_index, symbol)


def download_ticker_client_types_record(ticker_index: str):
    data = _extract_ticker_client_types_data(ticker_index)
    client_types_data_frame = pd.DataFrame(data, columns=[
        "date",
        "individual_buy_count", "corporate_buy_count",
        "individual_sell_count", "corporate_sell_count",
        "individual_buy_vol", "corporate_buy_vol",
        "individual_sell_vol", "corporate_sell_vol",
        "individual_buy_value", "corporate_buy_value",
        "individual_sell_value", "corporate_sell_value"
    ])
    for i in [
        "individual_buy_", "individual_sell_",
        "corporate_buy_", "corporate_sell_"
    ]:
        client_types_data_frame[f"{i}mean_price"] = (
                client_types_data_frame[f"{i}value"].astype(float) /
                client_types_data_frame[f"{i}vol"].astype(float)
        )
    client_types_data_frame["individual_ownership_change"] = (
            client_types_data_frame["corporate_sell_vol"].astype(float) -
            client_types_data_frame["corporate_buy_vol"].astype(float)
    )
    return client_types_data_frame


def _extract_ticker_client_types_data(ticker_index: str) -> List:
    url = TSE_CLIENT_TYPE_DATA_URL.format(ticker_index)
    with requests_retry_session() as session:
        response = session.get(url, timeout=5)
    data = response.text.split(";")
    data = [row.split(",") for row in data]
    return data


def to_arabic(string: str):
    return string.replace('ک', 'ك').replace('ی', 'ي').strip()


def get_symbol_id(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception("Sorry, tse server did not respond")

    symbol_full_info = response.text.split(';')[0].split(',')
    if (to_arabic(symbol_name) == symbol_full_info[0].strip()):
        return symbol_full_info[2]  # symbol id
    return None



import ast
import re
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class ShareholderData:
    shares: float
    percentage: float
    instrument_id: str
    name: str


def scrape_daily_info_page_for_shareholder_data(
    daily_info_page: str
) -> Optional[List[ShareholderData]]:
    """
    scrapes and returns shareholders data from a day info page like:
    http://cdn.tsetmc.com/Loader.aspx?ParTree=15131P&i=34557241988629814&d=20200201
    """
    shareholder_data_str = re.findall(
        'ShareHolderData=(.*?);', daily_info_page
    )
    if shareholder_data_str == "[]":
        return
    shareholders_data = ast.literal_eval(shareholder_data_str[0])
    return [
        ShareholderData(
            shares=shareholder_data[2],
            percentage=shareholder_data[3],
            instrument_id=shareholder_data[1],
            name=shareholder_data[5],
        ) for shareholder_data in shareholders_data
    ]


def scrape_daily_info_page_for_date(
    daily_info_page: str
) -> Optional[List[ShareholderData]]:
    daily_info_page_date = re.findall("DEven='(.*?)'", daily_info_page)
    if len(daily_info_page_date) != 1:
        return
    return daily_info_page_date[0]



import json
from typing import Set, Dict

from pytse_client import config

ticker_name_to_index_mapping = None


def tickers_dict() -> Dict:
    global ticker_name_to_index_mapping
    if ticker_name_to_index_mapping is None:
        with open(
                f"{config.pytse_dir}/data/symbols_name.json",
                "r",
                encoding="utf8"
        ) as symbols_name:
            ticker_name_to_index_mapping = json.load(symbols_name)
    return ticker_name_to_index_mapping


def get_ticker_index(ticker_symbol: str):
    return tickers_dict().get(ticker_symbol)


def all_symbols() -> Set:
    return set(tickers_dict().keys())


def append_symbol_to_file(symbol_id: str, symbol_name: str):
    new_symbol = {symbol_name.strip(): symbol_id.strip()}
    with open(
        f"{config.pytse_dir}/data/symbols_name.json",
        "r+",
        encoding="utf8"
    ) as file:
        data = json.load(file)
        data.update(new_symbol)
        file.seek(0)
        json.dump(data, file, ensure_ascii=False, indent=2)
